<?php

namespace App\Http\Controllers\Admin;

use App\Http\Requests\ClientRequest;
use Backpack\CRUD\app\Http\Controllers\CrudController;
use Backpack\CRUD\app\Library\CrudPanel\CrudPanelFacade as CRUD;
use Twilio\Rest\Client as TWL;
use Illuminate\Support\Facades\Log;
use Backpack\CRUD\app\Library\Widget;

/**
 * Class ClientCrudController
 * @package App\Http\Controllers\Admin
 * @property-read \Backpack\CRUD\app\Library\CrudPanel\CrudPanel $crud
 */
class ClientCrudController extends CrudController
{
    use \Backpack\CRUD\app\Http\Controllers\Operations\ListOperation;
    use \Backpack\CRUD\app\Http\Controllers\Operations\CreateOperation { store as traitStore; }
    use \Backpack\CRUD\app\Http\Controllers\Operations\UpdateOperation;
    use \Backpack\CRUD\app\Http\Controllers\Operations\DeleteOperation { destroy as traitDestroy; }
    use \Backpack\CRUD\app\Http\Controllers\Operations\ShowOperation;

    protected $_twilioClient;
    public function __construct(TWL $twilioClient)
    {
        parent::__construct();
        $this->_twilioClient = $twilioClient;
    }

    /**
     * Configure the CrudPanel object. Apply settings to all operations.
     *
     * @return void
     */
    public function setup()
    {
        CRUD::setModel(\App\Models\Client::class);
        CRUD::setRoute(config('backpack.base.route_prefix') . '/client');
        CRUD::setEntityNameStrings('client', 'clients');
    }

    /**
     * Define what happens when the List operation is loaded.
     *
     * @see  https://backpackforlaravel.com/docs/crud-operation-list-entries
     * @return void
     */
    protected function setupListOperation()
    {
        Widget::add()
    ->to('before_content')
    ->type('card')
    ->content(['header' => 'Some card title', // optional
        'body'   => 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Duis non mi nec orci euismod venenatis. Integer quis sapien et diam facilisis facilisis ultricies quis justo. Phasellus sem <b>turpis</b>, ornare quis aliquet ut, volutpat et lectus. Aliquam a egestas elit. <i>Nulla posuere</i>, sem et porttitor mollis, massa nibh sagittis nibh, id porttitor nibh turpis sed arcu.']);

        CRUD::column('name');
        $this->crud->query->withCount('trackingNumber'); // this will add a tags_count column to the results
        CRUD::addColumn([
           'name'      => 'trackingNumber', // name of relationship method in the model
           'type'      => 'relationship_count',
           'label'     => 'Tracking Numbers', // Table column heading
           'suffix'    => ' tracking numbers'
        ]);
        
        $this->crud->query->withCount('numbers'); // this will add a tags_count column to the results
        CRUD::addColumn([
           'name'      => 'numbers', // name of relationship method in the model
           'type'      => 'relationship_count',
           'label'     => 'Phone Numbers', // Table column heading
           'suffix'    => ' phone numbers'
        ]);
        
        CRUD::addColumn([
            'name'  => 'enabled',
            'label' => 'Status',
            'type'  => 'boolean',
            'options' => [0 => 'Inactive', 1 => 'Active']
        ]);
        
        /**
         * Columns can be defined using the fluent syntax or array syntax:
         * - CRUD::column('price')->type('number');
         * - CRUD::addColumn(['name' => 'price', 'type' => 'number']);
         */
    }

    /**
     * Define what happens when the Create operation is loaded.
     *
     * @see https://backpackforlaravel.com/docs/crud-operation-create
     * @return void
     */
    protected function setupCreateOperation()
    {
        CRUD::setValidation(ClientRequest::class);

        CRUD::field('name');
        CRUD::field('phone_number');
        CRUD::field('email');
        CRUD::addField([  
            'name'  => 'enabled',
            'type'  => 'switch',
            'label'    => 'Enable Tracking Numbers',
            // optional
            'color'    => 'primary', // May be any bootstrap color class or an hex color
            'onLabel' => '✓',
            'offLabel' => '✕',
        ]);

        /**
         * Fields can be defined using the fluent syntax or array syntax:
         * - CRUD::field('price')->type('number');
         * - CRUD::addField(['name' => 'price', 'type' => 'number']));
         */
    }

    /**
     * Define what happens when the Update operation is loaded.
     *
     * @see https://backpackforlaravel.com/docs/crud-operation-update
     * @return void
     */
    protected function setupUpdateOperation()
    {
        $this->setupCreateOperation();
    }

    /** SAVE CRUD AND CREATE TWILIO SUBACCOUNT **/
    public function store()
    {
      $data = $this->crud->getRequest()->request->all();

      if(array_key_exists('name',$data) && $data['name'] != '') {
        $account = $this->_twilioClient->api->v2010->accounts->create(["friendlyName" => $data['name']]);
        //$account = $this->_twilioClient->api->v2010->accounts("ACb2627066b91ed1c5af2ef308f0fa2e00")
        //                              ->update(["status" => "closed"]);

        $this->crud->getRequest()->request->add(['auth_token' => $account->authToken]);
        $this->crud->addField(['type' => 'hidden', 'name' => 'auth_token']);

        $this->crud->getRequest()->request->add(['sid' => $account->sid]);
        $this->crud->addField(['type' => 'hidden', 'name' => 'sid']);
        Log::info($account);
        if(!is_null($account)) {
          $this->_twilioClient = new TWL($account->sid, $account->authToken);
          $app = $this->_twilioClient->applications
                      ->create([
                                   "voiceMethod" => "POST",
                                   "voiceUrl" => "https://analytics.lucid.team/api/call-log",
                                   "statusCallback" => "https://analytics.lucid.team/api/call-log",
                                   "statusCallbackMethod" => "POST",
                                   "voiceCallerIdLookup" => "true",
                                   "friendlyName" => "Call Analytics"
                               ]
                      );
        Log::info($app);
          $this->crud->getRequest()->request->add(['app_sid' => $app->sid]);
          $this->crud->addField(['type' => 'hidden', 'name' => 'app_sid']);
        }
      }

        $response = $this->traitStore();

        return $response;
    }

    /** DELETE CRUD AND DELETE TWILIO SUBACCOUNT **/
    public function destroy($id)
    {
        $this->crud->hasAccessOrFail('delete');
        $client = $this->crud->getEntry($id);

        if(!is_null($client->app_sid)) {
          $this->_twilioClient->applications($client->app_sid)->delete();
        }

        if(!is_null($client->sid)) {
          $this->_twilioClient->api->v2010->accounts($client->sid)->update(["status" => "closed"]);
        }

        return $this->crud->delete($id);
    }
    
      // show whatever you want
    protected function setupShowOperation()
    {   
        $client = $this->crud->getCurrentEntry();
        $this->_twilioClient = new TWL($client->sid, $client->auth_token);
          
        $totalTMonth = $this->_twilioClient->usage->records->lastMonth->read(array(
                                "category" => "totalprice",
                            ));
        $totalLMonth = $this->_twilioClient->usage->records->thisMonth->read(array(
                                "category" => "totalprice",
                            ));
     
    Widget::add()
    ->to('before_content')
    ->type('card')
    ->content(['header' => $totalTMonth[0]->price, // optional
        'body'   => 'Cost (This Month)'
        ]);    
    Widget::add()
    ->to('before_content')
    ->type('card')
    ->content(['header' => $totalLMonth[0]->price, // optional
        'body'   => 'Cost (Last Month)'
        ]);      
        
        // MAYBE: do stuff before the autosetup
Widget::add()
    ->to('before_content')
    ->type('card')
    ->content(['header' => $this->crud->getCurrentEntry()->call_log->count(), // optional
        'body'   => 'Calls (This Month)'
        ]);

        // MAYBE: do stuff after the autosetup
    $id = $this->crud->getCurrentEntry()->id;
    CRUD::addColumn([
        'name'     => 'my_custom_html',
        'label'    => 'Website Tracking Code',
        'type'     => 'custom_html',
        'value'    => '<pre class="prettyprint">&lt;script type="text/javascript" src="https://analytics.lucid.team/api/client/'.$id.'/swap.js"&gt;&lt;/script&gt;
</pre>',
    
        // OPTIONALS
        // 'escaped' => true // echo using {{ }} instead of {!! !!}
    ]);   
    
    // automatically add the columns
        $this->autoSetupShowOperation();
    }
}
