<?php

namespace Backpack\DevTools\Generators;

use Backpack\DevTools\CustomFile;
use Backpack\DevTools\Models\Operation;
use Exception;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\File;
use Illuminate\Support\Str;

class OperationGenerator
{
    public const TYPE_GLOBAL = 'global';

    public const TYPE_LINE = 'line';

    public const TYPE_BULK = 'bulk';

    public const ACTION_MAKE_GET_REQUEST_TO_BACKPACK_FORM = 'make_get_request_to_backpack_form';

    public const ACTION_MAKE_GET_REQUEST_TO_CUSTOM_VIEW = 'make_get_request_to_custom_view';

    public const ACTION_CONFIRM_AND_MAKE_AJAX_CALL = 'confirm_and_make_ajax_call';

    public const ACTION_MAKE_AJAX_CALL = 'make_ajax_call';

    public const ACTION_OPEN_MODAL = 'open_modal';

    public const BUTTON_ACTIONS = [
        self::ACTION_MAKE_GET_REQUEST_TO_BACKPACK_FORM,
        self::ACTION_MAKE_GET_REQUEST_TO_CUSTOM_VIEW,
        self::ACTION_CONFIRM_AND_MAKE_AJAX_CALL,
        self::ACTION_MAKE_AJAX_CALL,
        self::ACTION_OPEN_MODAL,
    ];

    public const TYPES = [
        self::TYPE_GLOBAL,
        self::TYPE_LINE,
        self::TYPE_BULK,
    ];

    private const PATH_TO_STUBS = '/../Stubs/';

    private const SUFFIX = 'Operation.php';

    private const TEXT_HOLDER_OPERATION_NAME = '___to_be_replaced_with_operation_name___';

    private const TEXT_HOLDER_TRAIT_NAME = '___to_be_replaced_with_trait_operation___';

    private const TEXT_HOLDER_BUTTON_POSITION = '___to_be_replaced_with_button_position___';

    private const TEXT_HOLDER_BUTTON_LABEL = '___to_be_replaced_with_button_label___';

    private const TEXT_HOLDER_CONFIRMATION_MESSAGE = '___to_be_replaced_with_confirmation_message___';

    private const TEXT_HOLDER_NAMESPACE = '___to_be_replaced_with_namespace___';

    private const PLACEHOLDER_BUTTON = '___to_be_replaced_with_button___';

    private const PLACEHOLDER_ROUTE_GET = '___to_be_replaced_with_route_get___';

    private const PLACEHOLDER_ROUTE_POST = '___to_be_replaced_with_route_post___';

    private const PLACEHOLDER_METHOD_GET = '___to_be_replaced_with_method_get___';

    private const PLACEHOLDER_METHOD_POST = '___to_be_replaced_with_method_post___';

    private const PLACEHOLDER_ENABLE_BULK_ACTION = '___to_be_replaced_with_enable_bulk_action___';

    /**
     * Where to create the new operation.
     */
    private string $path;

    /**
     * Name given by user to the new operation.
     */
    private string $name;

    /**
     * We build this by using the path + name + self::suffix.
     */
    private string $fullPath;

    /**
     * We build this by using the path in snake format.
     */
    private string $pathNames;

    /**
     * Sample file content used to build the operation.
     */
    private string $template;

    /**
     * A list of controllers to add the new operation after creation.
     */
    private array $controllers;

    /**
     * The type of operation: global, line, bulk.
     */
    private string $type;

    /**
     * The message to display to admins in order to confirm the execution of the operation.
     * Only applies when $buttonLabel is with confirmation message.
     */
    private ?string $confirmationMessage;

    /**
     * Just the button label!
     */
    private string $buttonLabel;

    /**
     * Type of action when the button is clicked:
     *      - make a get request to a custom Backpack view
     *      - make a get request to a custom view
     *      - confirm and make an ajax call
     *      - make an ajax call
     *      - open a modal.
     */
    private string $buttonAction;

    public function __construct(Request $request)
    {
        $this->path = $request->input('path');
        $this->name = Str::of($request->input('name'))->studly();
        $this->pathNames = Str::of($this->name)->studly();
        $this->type = $request->input('type');
        $this->controllers = $request->input('controllers') ?? [];
        $this->buttonLabel = $request->input('button_label');
        $this->buttonAction = $request->input('button_action');
        $this->confirmationMessage = $request->input('confirmation_message');
    }

    /**
     * @throws Exception
     */
    public function generate(): ?int
    {
        $this->buildTemplate()
            ->buildRequiredViews()
            ->replaceStrings()
            ->save()
            ->addToControllers();

        return $this->getCreatedFileKey();
    }

    private function buildTemplate(): self
    {
        return $this->buildFullPath()->getTemplateContent()->buildOperationTrait();
    }

    public function buildRequiredViews(): self
    {
        return $this->buildAndSaveButtonView()->buildAndSaveCustomView();
    }

    private function buildAndSaveButtonView(): self
    {
        $directory = resource_path('views/vendor/backpack/crud/buttons/');
        $this->createDirectoryIfDoesntExist($directory);
        $path = $directory.lcfirst($this->pathNames).'.blade.php';

        // Check if View already exists
        if (File::exists($path)) {
            return $this;
        }

        $buttonViewContent = File::get(__DIR__.self::PATH_TO_STUBS.'views/button_'.$this->type.'/'.$this->buttonAction.'.stub');
        $buttonViewContent = Str::replace(self::TEXT_HOLDER_BUTTON_LABEL, $this->buttonLabel, $buttonViewContent);
        $buttonViewContent = Str::replace(self::TEXT_HOLDER_OPERATION_NAME, lcfirst($this->pathNames), $buttonViewContent);

        if (! empty($this->confirmationMessage)) {
            $buttonViewContent = Str::replace(
                self::TEXT_HOLDER_CONFIRMATION_MESSAGE,
                $this->confirmationMessage,
                $buttonViewContent,
            );
        }

        File::put($path, $buttonViewContent);

        return $this;
    }

    private function buildAndSaveCustomView(): self
    {
        if (! $this->shouldBuildView()) {
            return $this;
        }

        $directory = resource_path('views/vendor/backpack/crud/operations/');
        $this->createDirectoryIfDoesntExist($directory);
        $path = $directory.lcfirst($this->pathNames).'.blade.php';

        // Check if View already exists
        if (File::exists($path)) {
            return $this;
        }

        $methodType = $this->type === 'global' && $this->buttonAction !== self::ACTION_MAKE_GET_REQUEST_TO_CUSTOM_VIEW ? '_global' : '';
        $customViewContent = File::get(__DIR__.self::PATH_TO_STUBS.'views/'.$this->buttonAction.$methodType.'.stub');

        File::put($path, $customViewContent);

        return $this;
    }

    private function replaceStrings(): self
    {
        return $this->replaceNamespace()
            ->replaceTraitName()
            ->replaceButtonStrings()
            ->replacePathStrings();
    }

    private function buildOperationTrait(): self
    {
        $requiresGetRoute = $this->shouldBuildView();
        $withId = $this->type === self::TYPE_LINE ? '_with_id' : '';

        // Route GET
        $getRouteContent = $requiresGetRoute
            ? File::get(__DIR__.self::PATH_TO_STUBS.'inc/routes/get'.$withId.'.stub').'        '
            : '';
        $this->template = Str::replace(self::PLACEHOLDER_ROUTE_GET, $getRouteContent, $this->template);

        // Route POST
        $postRouteContent = File::get(__DIR__.self::PATH_TO_STUBS.'inc/routes/post'.$withId.'.stub');
        $this->template = Str::replace(self::PLACEHOLDER_ROUTE_POST, $postRouteContent, $this->template);

        // Method GET
        $getPath = $this->buttonAction !== self::ACTION_MAKE_GET_REQUEST_TO_CUSTOM_VIEW
            ? $withId
            : '_for_custom_view';
        $getMethodContent = $requiresGetRoute
            ? File::get(__DIR__.self::PATH_TO_STUBS.'inc/methods/get'.$getPath.'.stub').PHP_EOL.'    '
            : '';
        $this->template = Str::replace(self::PLACEHOLDER_METHOD_GET, $getMethodContent, $this->template);

        // Method POST
        $postMethodContent = File::get(__DIR__.self::PATH_TO_STUBS.'inc/methods/post'.$withId.'.stub');
        $this->template = Str::replace(self::PLACEHOLDER_METHOD_POST, $postMethodContent, $this->template);

        // Button
        $buttonContent = File::get(__DIR__.self::PATH_TO_STUBS.'inc/button.stub');
        $this->template = Str::replace(self::PLACEHOLDER_BUTTON, $buttonContent, $this->template);

        $enableBulk = $this->type === self::TYPE_BULK ? PHP_EOL.'            $this->crud->enableBulkActions();' : '';
        $this->template = Str::replace(self::PLACEHOLDER_ENABLE_BULK_ACTION, $enableBulk, $this->template);

        return $this;
    }

    private function buildFullPath(): self
    {
        $this->fullPath = $this->path.'/'.$this->name.self::SUFFIX;

        return $this;
    }

    /**
     * @throws Exception
     */
    private function save(): self
    {
        $this->createDirectoryIfDoesntExist($this->path);

        if (! File::put($this->fullPath, $this->template)) {
            throw new Exception('Operation was not generated.');
        }

        return $this;
    }

    private function getCreatedFileKey(): ?int
    {
        Operation::clearBootedModels();

        $item = Operation::query()->where('file_path', $this->fullPath)->first();

        return $item ? $item->getKey() : null;
    }

    private function getTemplateContent(): self
    {
        $this->template = File::get(__DIR__.self::PATH_TO_STUBS.'trait.stub');

        return $this;
    }

    private function getNamespace(): string
    {
        $namespace = 'App'.str_replace(app_path(), '', $this->path);

        return str_replace('/', '\\', $namespace);
    }

    private function replaceNamespace(): self
    {
        $this->template = Str::replace(self::TEXT_HOLDER_NAMESPACE, $this->getNamespace(), $this->template);

        return $this;
    }

    private function replaceTraitName(): self
    {
        $this->template = Str::replace(self::TEXT_HOLDER_TRAIT_NAME, $this->name, $this->template);

        return $this;
    }

    private function replaceButtonStrings(): self
    {
        switch ($this->type) {
            case self::TYPE_GLOBAL:
                $buttonPosition = 'top';
                break;
            case self::TYPE_LINE:
                $buttonPosition = 'line';
                break;
            case self::TYPE_BULK:
                $buttonPosition = 'bottom';
                break;
            default:
                // This should never be reached since it has been validated.
                // Note: It should not be reached since it was already validated!
                $buttonPosition = '';
        }

        $this->template = Str::replace(self::TEXT_HOLDER_BUTTON_POSITION, $buttonPosition, $this->template);

        return $this;
    }

    private function replacePathStrings(): self
    {
        $this->template = Str::replace(self::TEXT_HOLDER_OPERATION_NAME, lcfirst($this->pathNames), $this->template);

        return $this;
    }

    /**
     * We divide the Controller file content in three parts to be able to manipulate it:
     *   1- Content before Class definition — E.g. "Class MyController extends CrudController"
     *   2- Content before methods -> here is where we look for "use" statements to place at the end the new one
     *   3- Content methods.
     *
     * @return void
     */
    private function addToControllers(): void
    {
        foreach ($this->controllers ?? [] as $controllerPath) {
            // Load Controller file
            $controller = CustomFile::allFrom(config('backpack.devtools.paths.crud_controllers', []))
                ->filter(function ($file) use ($controllerPath) {
                    return $file->isClass() && $file->isCrudController() && $file->file_path === $controllerPath;
                })
                ->first();

            $classDefinition = $controller->class_name.' extends CrudController';

            // If Controller does not extend CrudController, we ignore it
            if (! Str::contains($controller->file_contents, $classDefinition)) {
                continue;
            }

            // 1- Get content before Class definition
            $beforeClassDefinitionSection = Str::before($controller->file_contents, $classDefinition).$classDefinition.PHP_EOL.'{';

            // 2- Get content after Class definition and make sure we are working after "{"
            $afterClassDefinitionSection = Str::after(Str::after($controller->file_contents, $classDefinition), '{');

            // 3- Get existing use statements and methods content
            $useStatementsSection = Str::before($afterClassDefinitionSection, 'function').'function';
            $methodsSection = Str::after($afterClassDefinitionSection, 'function');

            // Place use statement at the end of existing ones
            if (Str::contains($useStatementsSection, 'use ')) {
                $beforeLastUseStatement = Str::beforeLast($useStatementsSection, ';').';'.PHP_EOL;
                $beforeLastUseStatement .= $this->buildUseOperationForController();
                $afterLastUseStatement = Str::afterLast($useStatementsSection, ';');
                $useStatementsSection = $beforeLastUseStatement.$afterLastUseStatement;
            } else {
                // Add use statement right before the first method
                $useStatementsSection = PHP_EOL.$this->buildUseOperationForController().PHP_EOL.$useStatementsSection;
            }

            // Build file contents
            $controller->file_contents = $beforeClassDefinitionSection;
            $controller->file_contents .= $useStatementsSection;
            $controller->file_contents .= $methodsSection;

            // Rewrite Controller file
            File::put($controller->file_path, $controller->file_contents);
        }
    }

    private function buildUseOperationForController(): string
    {
        $str = '    use \\App'.str_replace(app_path(), '', $this->path);
        $str .= '\\'.Str::of($this->name)->studly().'Operation;';

        return str_replace('/', '\\', $str);
    }

    private function shouldBuildView(): bool
    {
        return $this->buttonAction === self::ACTION_MAKE_GET_REQUEST_TO_BACKPACK_FORM
            || $this->buttonAction === self::ACTION_MAKE_GET_REQUEST_TO_CUSTOM_VIEW
            || $this->buttonAction === self::ACTION_OPEN_MODAL;
    }

    private function createDirectoryIfDoesntExist(string $path)
    {
        if (! File::isDirectory($path)) {
            File::makeDirectory($path, 0777, true);
        }
    }
}
