<?php

namespace App\Http\Controllers\Admin;

use App\Http\Requests\ClientRequest;
use Backpack\CRUD\app\Http\Controllers\CrudController;
use Backpack\CRUD\app\Library\CrudPanel\CrudPanelFacade as CRUD;
use Twilio\Rest\Client as TWL;
use Illuminate\Support\Facades\Log;

/**
 * Class ClientCrudController
 * @package App\Http\Controllers\Admin
 * @property-read \Backpack\CRUD\app\Library\CrudPanel\CrudPanel $crud
 */
class ClientCrudController extends CrudController
{
    use \Backpack\CRUD\app\Http\Controllers\Operations\ListOperation;
    use \Backpack\CRUD\app\Http\Controllers\Operations\CreateOperation { store as traitStore; }
    use \Backpack\CRUD\app\Http\Controllers\Operations\UpdateOperation;
    use \Backpack\CRUD\app\Http\Controllers\Operations\DeleteOperation { destroy as traitDestroy; }
    use \Backpack\CRUD\app\Http\Controllers\Operations\ShowOperation;

    protected $_twilioClient;
    public function __construct(TWL $twilioClient)
    {
        parent::__construct();
        $this->_twilioClient = $twilioClient;
    }

    /**
     * Configure the CrudPanel object. Apply settings to all operations.
     *
     * @return void
     */
    public function setup()
    {
        CRUD::setModel(\App\Models\Client::class);
        CRUD::setRoute(config('backpack.base.route_prefix') . '/client');
        CRUD::setEntityNameStrings('client', 'clients');
    }

    /**
     * Define what happens when the List operation is loaded.
     *
     * @see  https://backpackforlaravel.com/docs/crud-operation-list-entries
     * @return void
     */
    protected function setupListOperation()
    {
        CRUD::column('name');
        CRUD::column('phone_number');
        CRUD::column('email');
        $this->crud->query->withCount('trackingNumber'); // this will add a tags_count column to the results
        CRUD::addColumn([
           'name'      => 'trackingNumber', // name of relationship method in the model
           'type'      => 'relationship_count',
           'label'     => 'Tracking Numbers', // Table column heading
           'suffix'    => ' tracking numbers'
        ]);
        /**
         * Columns can be defined using the fluent syntax or array syntax:
         * - CRUD::column('price')->type('number');
         * - CRUD::addColumn(['name' => 'price', 'type' => 'number']);
         */
    }

    /**
     * Define what happens when the Create operation is loaded.
     *
     * @see https://backpackforlaravel.com/docs/crud-operation-create
     * @return void
     */
    protected function setupCreateOperation()
    {
        CRUD::setValidation(ClientRequest::class);

        CRUD::field('name');
        CRUD::field('phone_number');
        CRUD::field('email');

        /**
         * Fields can be defined using the fluent syntax or array syntax:
         * - CRUD::field('price')->type('number');
         * - CRUD::addField(['name' => 'price', 'type' => 'number']));
         */
    }

    /**
     * Define what happens when the Update operation is loaded.
     *
     * @see https://backpackforlaravel.com/docs/crud-operation-update
     * @return void
     */
    protected function setupUpdateOperation()
    {
        $this->setupCreateOperation();
    }

    /** SAVE CRUD AND CREATE TWILIO SUBACCOUNT **/
    public function store()
    {
      $data = $this->crud->getRequest()->request->all();

      if(array_key_exists('name',$data) && $data['name'] != '') {
        $account = $this->_twilioClient->api->v2010->accounts->create(["friendlyName" => $data['name']]);
        //$account = $this->_twilioClient->api->v2010->accounts("ACb2627066b91ed1c5af2ef308f0fa2e00")
        //                              ->update(["status" => "closed"]);

        $this->crud->getRequest()->request->add(['auth_token' => $account->authToken]);
        $this->crud->addField(['type' => 'hidden', 'name' => 'auth_token']);

        $this->crud->getRequest()->request->add(['sid' => $account->sid]);
        $this->crud->addField(['type' => 'hidden', 'name' => 'sid']);
Log::info($account);
        if(!is_null($account)) {
          $this->_twilioClient = new TWL($account->sid, $account->authToken);
          $app = $this->_twilioClient->applications
                      ->create([
                                   "voiceMethod" => "POST",
                                   "voiceUrl" => "https://calls2.lucidadvertising.com/api/call-log",
                                   "statusCallback" => "https://calls2.lucidadvertising.com/api/call-log",
                                   "statusCallbackMethod" => "POST",
                                   "friendlyName" => "Call Analytics"
                               ]
                      );
        Log::info($app);
          $this->crud->getRequest()->request->add(['app_sid' => $app->sid]);
          $this->crud->addField(['type' => 'hidden', 'name' => 'app_sid']);
        }
      }

        $response = $this->traitStore();

        return $response;
    }

    /** DELETE CRUD AND DELETE TWILIO SUBACCOUNT **/
    public function destroy($id)
    {
        $this->crud->hasAccessOrFail('delete');
        $client = $this->crud->getEntry($id);

        if(!is_null($client->app_sid)) {
          $this->_twilioClient->applications($client->app_sid)->delete();
        }

        if(!is_null($client->sid)) {
          $this->_twilioClient->api->v2010->accounts($client->sid)->update(["status" => "closed"]);
        }

        return $this->crud->delete($id);
    }
}
