<?php

namespace App\Http\Controllers\Admin;

use App\Http\Requests\LogRequest;
use Backpack\CRUD\app\Http\Controllers\CrudController;
use Backpack\CRUD\app\Library\CrudPanel\CrudPanelFacade as CRUD;

/**
 * Class LogCrudController
 * @package App\Http\Controllers\Admin
 * @property-read \Backpack\CRUD\app\Library\CrudPanel\CrudPanel $crud
 */
class LogCrudController extends CrudController
{
    use \Backpack\CRUD\app\Http\Controllers\Operations\ListOperation;
    use \Backpack\CRUD\app\Http\Controllers\Operations\CreateOperation;
    use \Backpack\CRUD\app\Http\Controllers\Operations\UpdateOperation;
    use \Backpack\CRUD\app\Http\Controllers\Operations\DeleteOperation;
    use \Backpack\CRUD\app\Http\Controllers\Operations\ShowOperation;

    /**
     * Configure the CrudPanel object. Apply settings to all operations.
     *
     * @return void
     */
    public function setup()
    {
        CRUD::setModel(\App\Models\Log::class);
        CRUD::setRoute(config('backpack.base.route_prefix') . '/log');
        CRUD::setEntityNameStrings('Call', 'Call Logs');
    }

    /**
     * Define what happens when the List operation is loaded.
     *
     * @see  https://backpackforlaravel.com/docs/crud-operation-list-entries
     * @return void
     */
    protected function setupListOperation()
    {
        CRUD::column('duration');
        CRUD::column('status');
        CRUD::column('created_at');

        /**
         * Columns can be defined using the fluent syntax or array syntax:
         * - CRUD::column('price')->type('number');
         * - CRUD::addColumn(['name' => 'price', 'type' => 'number']);
         */
         CRUD::addColumn([
           'label'     => 'Tracking Number', // Table column heading
           'type'      => 'select',
           'name'      => 'number_id', // the column that contains the ID of that connected entity;
           'entity'    => 'trackingNumber', // the method that defines the relationship in your Model
           'attribute' => 'name', // foreign key attribute that is shown to user
           'model'     => "App\Models\TrackingNumber", // foreign key model
         ]);
         CRUD::addColumn([
           'label'     => 'Caller', // Table column heading
           'type'      => 'select',
           'name'      => 'caller_id', // the column that contains the ID of that connected entity;
           'entity'    => 'caller', // the method that defines the relationship in your Model
           'attribute' => 'name', // foreign key attribute that is shown to user
           'model'     => "App\Models\Caller", // foreign key model
         ]);
    }

    /**
     * Define what happens when the Create operation is loaded.
     *
     * @see https://backpackforlaravel.com/docs/crud-operation-create
     * @return void
     */
    protected function setupCreateOperation()
    {
        CRUD::setValidation(LogRequest::class);

        CRUD::field('sid');
        CRUD::field('duration');
        CRUD::field('status');

        /**
         * Fields can be defined using the fluent syntax or array syntax:
         * - CRUD::field('price')->type('number');
         * - CRUD::addField(['name' => 'price', 'type' => 'number']));
         */

         CRUD::addField([
                    // 1-n relationship
                    'label'     => 'Tracking Number', // Table column heading
                    'type'      => 'select_grouped',
                    'name'      => 'tracking_number_id', // the column that contains the ID of that connected entity;
                    'entity'    => 'trackingNumber', // the method that defines the relationship in your Model
                    'attribute' => 'name', // foreign key attribute that is shown to user
                    'model'     => "App\Models\TrackingNumber", // foreign key model
                    'group_by'  => 'client', // the relationship to entity you want to use for grouping
                    'group_by_attribute' => 'name', // the attribute on related model, that you want shown
                    'group_by_relationship_back' => 'trackingNumber', // relationship from related model back to this model
                    'options'   => (function ($query) {
                 return $query->orderBy('name', 'ASC')->get();
               }),
             ]);

             CRUD::addField([
                        // 1-n relationship
                        'label'     => 'Caller', // Table column heading
                        'type'      => 'select',
                        'name'      => 'caller_id', // the column that contains the ID of that connected entity;
                        'entity'    => 'caller', // the method that defines the relationship in your Model
                        'attribute' => 'name', // foreign key attribute that is shown to user
                        'model'     => "App\Models\Caller", // foreign key model
                        'options'   => (function ($query) {
                     return $query->orderBy('name', 'ASC')->get();
                   }),
                 ]);

    }

    /**
     * Define what happens when the Update operation is loaded.
     *
     * @see https://backpackforlaravel.com/docs/crud-operation-update
     * @return void
     */
    protected function setupUpdateOperation()
    {
        $this->setupCreateOperation();
    }
}
