<?php

namespace App\Http\Controllers\Admin;

use App\Http\Requests\TrackingNumberRequest;
use Backpack\CRUD\app\Http\Controllers\CrudController;
use Backpack\CRUD\app\Library\CrudPanel\CrudPanelFacade as CRUD;
use Twilio\Rest\Client as TWL;
use App\Models\BuyNumber;
use App\Models\Client;
use Illuminate\Support\Facades\Log;

/**
 * Class TrackingNumberCrudController
 * @package App\Http\Controllers\Admin
 * @property-read \Backpack\CRUD\app\Library\CrudPanel\CrudPanel $crud
 */
class TrackingNumberCrudController extends CrudController
{
    use \Backpack\CRUD\app\Http\Controllers\Operations\ListOperation;
    use \Backpack\CRUD\app\Http\Controllers\Operations\CreateOperation { store as traitStore; }
    use \Backpack\CRUD\app\Http\Controllers\Operations\UpdateOperation;
    use \Backpack\CRUD\app\Http\Controllers\Operations\DeleteOperation;
    use \Backpack\CRUD\app\Http\Controllers\Operations\ShowOperation;

    protected $_twilioClient;
    public function __construct(TWL $twilioClient)
    {
        parent::__construct();
        $this->_appSid = config('app.twilio_app_sid');
        $this->_twilioClient = $twilioClient;
    }

    /**
     * Configure the CrudPanel object. Apply settings to all operations.
     *
     * @return void
     */
    public function setup()
    {
        CRUD::setModel(\App\Models\TrackingNumber::class);
        CRUD::setRoute(config('backpack.base.route_prefix') . '/tracking-number');
        CRUD::setEntityNameStrings('tracking number', 'tracking numbers');
    }

    /**
     * Define what happens when the List operation is loaded.
     *
     * @see  https://backpackforlaravel.com/docs/crud-operation-list-entries
     * @return void
     */
    protected function setupListOperation()
    {
        CRUD::addColumn([
          'label'     => 'Client', // Table column heading
          'type'      => 'select',
          'name'      => 'client_id', // the column that contains the ID of that connected entity;
          'entity'    => 'client', // the method that defines the relationship in your Model
          'attribute' => 'name', // foreign key attribute that is shown to user
          'model'     => "App\Models\Client", // foreign key model
        ]);
        CRUD::column('name');
        CRUD::column('friendly_name');
        CRUD::addColumn('created_at');


        /**
         * Columns can be defined using the fluent syntax or array syntax:
         * - CRUD::column('price')->type('number');
         * - CRUD::addColumn(['name' => 'price', 'type' => 'number']);
         */
    }

    /**
     * Define what happens when the Create operation is loaded.
     *
     * @see https://backpackforlaravel.com/docs/crud-operation-create
     * @return void
     */
    protected function setupCreateOperation()
    {
        CRUD::setValidation(TrackingNumberRequest::class);

        CRUD::field('name');

        CRUD::addField([
          'label'       => "Tracking Number", // Table column heading
          'type'        => "select2_from_ajax",
          'name'        => 'number_id', // the column that contains the ID of that connected entity
          'entity'      => 'buyNumber', // the method that defines the relationship in your Model
          'attribute'   => "number", // foreign key attribute that is shown to user
          'placeholder' => "Select a number",
          'minimum_input_length' => 3,
          'include_all_form_fields' => true,
          'data_source' => url("api/buy-number"), // url to controller search function (with /{id} should return model)
          'model'     => "App\Models\BuyNumber", // foreign key model
        ]);


        CRUD::addField([
          'label'     => 'Client', // Table column heading
          'type'      => 'select',
          'name'      => 'client_id', // the column that contains the ID of that connected entity;
          'entity'    => 'client', // the method that defines the relationship in your Model
          'attribute' => 'name', // foreign key attribute that is shown to user
          'model'     => "App\Models\Client", // foreign key model
          'options'   => (function ($query) {
            return $query->orderBy('name', 'ASC')->get();
          }),
        ]);
    }

    /**
     * Define what happens when the Update operation is loaded.
     *
     * @see https://backpackforlaravel.com/docs/crud-operation-update
     * @return void
     */
    protected function setupUpdateOperation()
    {
        $this->setupCreateOperation();
    }

    public function store()
    {

      $data = $this->crud->getRequest()->request->all();
      Log::info($this->_appSid);

      if(array_key_exists('number_id',$data)) {
        $number = BuyNumber::where('number', $data['number_id'])->first();
        $this->crud->getRequest()->request->add(['friendly_name' => $number->friendly_name]);
        $this->crud->addField(['type' => 'hidden', 'name' => 'friendly_name']);

        if(!is_null($data['client_id'])) {
          $client = Client::where('id', $data['client_id'])->first();
          $this->_twilioClient = new TWL($client->sid, $client->auth_token);
        }

        try {
          $bought_number = $this->_twilioClient->incomingPhoneNumbers
              ->create(
                  [
                      "phoneNumber" => $number->number,
                      "friendlyName" => $number->friendly_name,
                      "voiceApplicationSid" => $client->app_sid,
                      "voiceCallerIdLookup" => true
                  ]
              );
          $this->crud->getRequest()->request->add(['number_sid' => $bought_number->sid]);
          $this->crud->addField(['type' => 'hidden', 'name' => 'number_sid']);
        }
        catch(Exception $e) {
          Log::info($e);
        }
      }

      $response = $this->traitStore();

      BuyNumber::truncate();

      return $response;
    }

    public function destroy($id)
    {
        $this->crud->hasAccessOrFail('delete');
        $tracking_number =$this->crud->getEntry($id);
        Log::info($tracking_number);

        if(!is_null($tracking_number->client)) {
          $this->_twilioClient = new TWL($tracking_number->client->sid, $tracking_number->client->auth_token);
        }
        if(!is_null($tracking_number->number_sid)){
          $this->_twilioClient->incomingPhoneNumbers($tracking_number->number_sid)->delete();
        }


        return $this->crud->delete($id);
    }

}
